# -*- coding: utf-8 -*-
"""
A Translation module.

You can translate text using this module.
"""
import json
import random
import typing
from abc import ABC, abstractmethod

import httpcore
import httpx
from httpx import Timeout

from idioma import utils
from idioma.constants import (
    DEFAULT_CLIENT_SERVICE_URLS,
    DEFAULT_FALLBACK_SERVICE_URLS,
    DEFAULT_USER_AGENT, LANGCODES, LANGUAGES, SPECIAL_CASES,
    DEFAULT_RAISE_EXCEPTION, DUMMY_DATA
)
from idioma.exceptions import EmptyTranslationData
from idioma.gtoken import TokenAcquirer
from idioma.models import Translated, Detected, TranslatedPart
from idioma.retry_transport import RetryTransport


class BaseTranslator(ABC):
    EXCLUDES = ('en', 'ca', 'fr')
    RPC_ID = 'MkEWBc'
    POST_PARAMS = {
        'rpcids': RPC_ID,
        'bl': 'boq_translate-webserver_20201207.13_p0',
        'soc-app': 1,
        'soc-platform': 1,
        'soc-device': 1,
        'rt': 'c',
    }
    """Google Translate ajax API implementation class

    You have to create an instance of Translator to use this API

    :param service_urls: google translate url list. URLs will be used randomly.
                         For example ``['translate.google.com', 'translate.google.co.kr']``
                         To preferably use the non webapp api, service url should be translate.googleapis.com
    :type service_urls: a sequence of strings

    :param user_agent: the User-Agent header to send when making requests.
    :type user_agent: :class:`str`

    :param proxies: proxies configuration.
                    Dictionary mapping protocol or protocol and host to the URL of the proxy
                    For example ``{'http': 'foo.bar:3128', 'http://host.name': 'foo.bar:4012'}``
    :type proxies: dictionary

    :param timeout: Definition of timeout for httpx library.
                    Will be used for every request.
    :type timeout: number or a double of numbers
    :param proxies: proxies configuration.
                    Dictionary mapping protocol or protocol and host to the URL of the proxy
                    For example ``{'http': 'foo.bar:3128', 'http://host.name': 'foo.bar:4012'}``
    :param raise_exception: if `True` then raise exception if smth will go wrong
    :param http2: whether to use HTTP2 (default: True)
    :param use_fallback: use a fallback method
    :type raise_exception: boolean
    """

    def __init__(self, service_urls=DEFAULT_CLIENT_SERVICE_URLS,
                 user_agent=DEFAULT_USER_AGENT,
                 raise_exception=DEFAULT_RAISE_EXCEPTION,
                 proxies: typing.Dict[str, httpcore.AsyncHTTPProxy] = None,
                 timeout: Timeout = 30,
                 http2=True,
                 use_fallback=False):

        # for aenter
        self.http2 = http2
        self.proxies = proxies
        self.user_agent = user_agent
        self.timeout = timeout

        self.client = self._create_client()

        self.client.headers.update({
            'User-Agent': user_agent,
            'Referer': 'https://translate.google.com',
        })

        if timeout is not None:
            self.client.timeout = timeout

        if use_fallback:
            self.service_urls = DEFAULT_FALLBACK_SERVICE_URLS
            self.client_type = 'gtx'
            pass
        else:
            # default way of working: use the defined values from user app
            self.service_urls = service_urls
            self.client_type = 'tw-ob'
            self.token_acquirer = self._get_token_acquirer()

        self.raise_exception = raise_exception

    def _create_client(self):
        retry_transport = RetryTransport(
            wrapped_transport=httpx.HTTPTransport(retries=10),
            retryable_methods={'GET', 'POST'},
            # Google would take too "sorry" page sometimes with 302
            retry_status_codes={302}
        )
        return httpx.Client(
            http2=self.http2,
            proxies=self.proxies,
            transport=retry_transport,
            timeout=self.timeout
        )

    def _get_token_acquirer(self):
        return TokenAcquirer(client=self.client, host=self.service_urls[0])

    @classmethod
    def _build_rpc_request(cls, text: str, dest: str, src: str):
        return json.dumps([[
            [
                cls.RPC_ID,
                json.dumps([[text, src, dest, True], [None]],
                           separators=(',', ':')),
                None,
                'generic',
            ],
        ]], separators=(',', ':'))

    def _pick_service_url(self):
        if len(self.service_urls) == 1:
            return self.service_urls[0]
        return random.choice(self.service_urls)

    @staticmethod
    def _parse_extra_data(data):
        response_parts_name_mapping = {
            0: 'translation',
            1: 'all-translations',
            2: 'original-language',
            5: 'possible-translations',
            6: 'confidence',
            7: 'possible-mistakes',
            8: 'language',
            11: 'synonyms',
            12: 'definitions',
            13: 'examples',
            14: 'see-also',
        }

        extra = {}

        for index, category in response_parts_name_mapping.items():
            extra[category] = data[index] if (
                    index < len(data) and data[index]) else None

        return extra

    @staticmethod
    def validate_normalize_src_dest(src, dest):
        dest = dest.lower().split('_', 1)[0]
        if src:
            src = src.lower().split('_', 1)[0]
        else:
            src = 'auto'

        if src != 'auto' and src not in LANGUAGES:
            if src in SPECIAL_CASES:
                src = SPECIAL_CASES[src]
            elif src in LANGCODES:
                src = LANGCODES[src]
            else:
                raise ValueError('invalid source language')

        if dest not in LANGUAGES:
            if dest in SPECIAL_CASES:
                dest = SPECIAL_CASES[dest]
            elif dest in LANGCODES:
                dest = LANGCODES[dest]
            else:
                raise ValueError('invalid destination language')
        return src, dest

    def parse_response(self, src, dest, origin, data, response: httpx.Response):
        token_found = False
        square_bracket_counts = [0, 0]
        resp = ''
        for line in data.split('\n'):
            token_found = token_found or f'"{self.RPC_ID}"' in line[:30]
            if not token_found:
                continue

            is_in_string = False
            for index, char in enumerate(line):
                if char == '\"' and line[max(0, index - 1)] != '\\':
                    is_in_string = not is_in_string
                if not is_in_string:
                    if char == '[':
                        square_bracket_counts[0] += 1
                    elif char == ']':
                        square_bracket_counts[1] += 1

            resp += line
            if square_bracket_counts[0] == square_bracket_counts[1]:
                break

        data = json.loads(resp)
        if data[0][2] is None:
            # Failed translation
            raise EmptyTranslationData(f"Empty translation data: {data}")
        parsed = json.loads(data[0][2])
        # not sure
        should_spacing = parsed[1][0][0][3]
        translated_parts = list(map(lambda part: TranslatedPart(part[0],
                                                                part[1] if len(
                                                                    part) >= 2 else []),
                                    parsed[1][0][0][5]))
        translated = (' ' if should_spacing else '').join(
            map(lambda part: part.text, translated_parts))

        if src == 'auto':
            try:
                src = parsed[2]
            except:
                pass
        if src == 'auto':
            try:
                src = parsed[0][2]
            except:
                pass

        # currently not available
        confidence = None

        origin_pronunciation = None
        try:
            origin_pronunciation = parsed[0][0]
        except:
            pass

        pronunciation = None
        try:
            pronunciation = parsed[1][0][0][1]
        except:
            pass

        extra_data = {
            'confidence': confidence,
            'parts': translated_parts,
            'origin_pronunciation': origin_pronunciation,
            'parsed': parsed,
        }
        result = Translated(src=src, dest=dest, origin=origin,
                            text=translated, pronunciation=pronunciation,
                            parts=translated_parts,
                            extra_data=extra_data,
                            response=response)
        return result

    @abstractmethod
    def translate(self, text: str, dest='en', src=None):
        pass

    def extract_source_language_and_confidence(self, data, response):
        # actual source language that will be recognized by Google Translator when the
        # src passed is equal to auto.
        src = ''
        confidence = 0.0
        try:
            if len(data[8][0]) > 1:
                src = data[8][0]
                confidence = data[8][-2]
            else:
                src = ''.join(data[8][0])
                confidence = data[8][-2][0]
        except Exception:  # pragma: nocover
            pass
        result = Detected(lang=src, confidence=confidence, response=response)

        return result

    def parse_legacy_response(self, data, src, dest, origin, response):
        # this code will be updated when the format is changed.
        translated = ''.join([d[0] if d[0] else '' for d in data[0]])

        extra_data = self._parse_extra_data(data)

        # actual source language that will be recognized by Google Translator when the
        # src passed is equal to auto.
        try:
            src = data[2]
        except Exception:  # pragma: nocover
            pass

        pron = origin
        try:
            pron = data[0][1][-2]
        except Exception:  # pragma: nocover
            pass

        if pron is None:
            try:
                pron = data[0][1][2]
            except:  # pragma: nocover
                pass

        if dest in self.EXCLUDES and pron == origin:
            pron = translated

        # put final values into a new Translated object
        result = Translated(src=src, dest=dest, origin=origin,
                            text=translated, pronunciation=pron,
                            extra_data=extra_data,
                            parts=[],
                            response=response)

        return result

    @abstractmethod
    def detect(self, text: str):
        pass

    @abstractmethod
    def detect_legacy(self, text, **kwargs):
        pass

    def handle_legacy_translate_response(self, response, text):
        if response.status_code == 200:
            data = utils.format_json(response.text)
            return data, response

        if self.raise_exception:
            raise Exception('Unexpected status code "{}" from {}'.format(
                response.status_code, self.service_urls))

        DUMMY_DATA[0][0][0] = text
        return DUMMY_DATA, response



from idioma import Translator


def test_translate_thai_hello():
    translator = Translator()
    assert translator.translate('안녕하세요').text == 'hello'


def test_translate_thai_hello_src_none_auto():
    translator = Translator()
    assert translator.translate('안녕하세요', src=None).text == 'hello'


def test_translate_thai_lang_detect():
    translator = Translator()
    assert translator.detect('안녕하세요').lang == 'ko'


def test_translate_legacy_russian_lang_hello():
    translator = Translator()
    assert translator.translate_legacy(
        'Привет', src='ru', dest='en').text == 'Hello'


def test_translate_russian_lang_hello_evade_rate_limit(capsys):
    translator = Translator(raise_exception=True)
    for i in range(300):
        translated = translator.translate(
            f'Привет #{i}', src='ru', dest='en').text
        print(f"Translated: {translated}")
        assert translated in [f'Hello #{i}', f'Hi #{i}', f'Hi: #{i}', f"Hi: {i}"]



"""Free Google Translate API for Python. Translates totally free of charge."""
__all__ = ['Translator', 'AsyncTranslator']
__version__ = '0.0.8'

from idioma.async_translator import AsyncTranslator
from idioma.constants import LANGCODES, LANGUAGES  # noqa
from idioma.translator import Translator



DEFAULT_USER_AGENT = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)'

DEFAULT_CLIENT_SERVICE_URLS = (
    'translate.google.com',
)

DEFAULT_FALLBACK_SERVICE_URLS = (
    'translate.googleapis.com',
)

DEFAULT_SERVICE_URLS = (
    'translate.google.ac', 'translate.google.ad', 'translate.google.ae',
    'translate.google.al', 'translate.google.am', 'translate.google.as',
    'translate.google.at', 'translate.google.az', 'translate.google.ba',
    'translate.google.be', 'translate.google.bf', 'translate.google.bg',
    'translate.google.bi', 'translate.google.bj', 'translate.google.bs',
    'translate.google.bt', 'translate.google.by', 'translate.google.ca',
    'translate.google.cat', 'translate.google.cc', 'translate.google.cd',
    'translate.google.cf', 'translate.google.cg', 'translate.google.ch',
    'translate.google.ci', 'translate.google.cl', 'translate.google.cm',
    'translate.google.cn', 'translate.google.co.ao', 'translate.google.co.bw',
    'translate.google.co.ck', 'translate.google.co.cr',
    'translate.google.co.id',
    'translate.google.co.il', 'translate.google.co.in',
    'translate.google.co.jp',
    'translate.google.co.ke', 'translate.google.co.kr',
    'translate.google.co.ls',
    'translate.google.co.ma', 'translate.google.co.mz',
    'translate.google.co.nz',
    'translate.google.co.th', 'translate.google.co.tz',
    'translate.google.co.ug',
    'translate.google.co.uk', 'translate.google.co.uz',
    'translate.google.co.ve',
    'translate.google.co.vi', 'translate.google.co.za',
    'translate.google.co.zm',
    'translate.google.co.zw', 'translate.google.com.af',
    'translate.google.com.ag',
    'translate.google.com.ai', 'translate.google.com.ar',
    'translate.google.com.au',
    'translate.google.com.bd', 'translate.google.com.bh',
    'translate.google.com.bn',
    'translate.google.com.bo', 'translate.google.com.br',
    'translate.google.com.bz',
    'translate.google.com.co', 'translate.google.com.cu',
    'translate.google.com.cy',
    'translate.google.com.do', 'translate.google.com.ec',
    'translate.google.com.eg',
    'translate.google.com.et', 'translate.google.com.fj',
    'translate.google.com.gh',
    'translate.google.com.gi', 'translate.google.com.gt',
    'translate.google.com.hk',
    'translate.google.com.jm', 'translate.google.com.kh',
    'translate.google.com.kw',
    'translate.google.com.lb', 'translate.google.com.ly',
    'translate.google.com.mm',
    'translate.google.com.mt', 'translate.google.com.mx',
    'translate.google.com.my',
    'translate.google.com.na', 'translate.google.com.ng',
    'translate.google.com.ni',
    'translate.google.com.np', 'translate.google.com.om',
    'translate.google.com.pa',
    'translate.google.com.pe', 'translate.google.com.pg',
    'translate.google.com.ph',
    'translate.google.com.pk', 'translate.google.com.pr',
    'translate.google.com.py',
    'translate.google.com.qa', 'translate.google.com.sa',
    'translate.google.com.sb',
    'translate.google.com.sg', 'translate.google.com.sl',
    'translate.google.com.sv',
    'translate.google.com.tj', 'translate.google.com.tr',
    'translate.google.com.tw',
    'translate.google.com.ua', 'translate.google.com.uy',
    'translate.google.com.vc',
    'translate.google.com.vn', 'translate.google.com', 'translate.google.cv',
    'translate.google.cz', 'translate.google.de', 'translate.google.dj',
    'translate.google.dk', 'translate.google.dm', 'translate.google.dz',
    'translate.google.ee', 'translate.google.es', 'translate.google.eu',
    'translate.google.fi', 'translate.google.fm', 'translate.google.fr',
    'translate.google.ga', 'translate.google.ge', 'translate.google.gf',
    'translate.google.gg', 'translate.google.gl', 'translate.google.gm',
    'translate.google.gp', 'translate.google.gr', 'translate.google.gy',
    'translate.google.hn', 'translate.google.hr', 'translate.google.ht',
    'translate.google.hu', 'translate.google.ie', 'translate.google.im',
    'translate.google.io', 'translate.google.iq', 'translate.google.is',
    'translate.google.it', 'translate.google.je', 'translate.google.jo',
    'translate.google.kg', 'translate.google.ki', 'translate.google.kz',
    'translate.google.la', 'translate.google.li', 'translate.google.lk',
    'translate.google.lt', 'translate.google.lu', 'translate.google.lv',
    'translate.google.md', 'translate.google.me', 'translate.google.mg',
    'translate.google.mk', 'translate.google.ml', 'translate.google.mn',
    'translate.google.ms', 'translate.google.mu', 'translate.google.mv',
    'translate.google.mw', 'translate.google.ne', 'translate.google.nf',
    'translate.google.nl', 'translate.google.no', 'translate.google.nr',
    'translate.google.nu', 'translate.google.pl', 'translate.google.pn',
    'translate.google.ps', 'translate.google.pt', 'translate.google.ro',
    'translate.google.rs', 'translate.google.ru', 'translate.google.rw',
    'translate.google.sc', 'translate.google.se', 'translate.google.sh',
    'translate.google.si', 'translate.google.sk', 'translate.google.sm',
    'translate.google.sn', 'translate.google.so', 'translate.google.sr',
    'translate.google.st', 'translate.google.td', 'translate.google.tg',
    'translate.google.tk', 'translate.google.tl', 'translate.google.tm',
    'translate.google.tn', 'translate.google.to', 'translate.google.tt',
    'translate.google.us', 'translate.google.vg', 'translate.google.vu',
    'translate.google.ws')

SPECIAL_CASES = {
    'ee': 'et',
}

LANGUAGES = {
    'af': 'afrikaans',
    'sq': 'albanian',
    'am': 'amharic',
    'ar': 'arabic',
    'hy': 'armenian',
    'az': 'azerbaijani',
    'eu': 'basque',
    'be': 'belarusian',
    'bn': 'bengali',
    'bs': 'bosnian',
    'bg': 'bulgarian',
    'ca': 'catalan',
    'ceb': 'cebuano',
    'ny': 'chichewa',
    'zh-cn': 'chinese (simplified)',
    'zh-tw': 'chinese (traditional)',
    'co': 'corsican',
    'hr': 'croatian',
    'cs': 'czech',
    'da': 'danish',
    'nl': 'dutch',
    'en': 'english',
    'eo': 'esperanto',
    'et': 'estonian',
    'tl': 'filipino',
    'fi': 'finnish',
    'fr': 'french',
    'fy': 'frisian',
    'gl': 'galician',
    'ka': 'georgian',
    'de': 'german',
    'el': 'greek',
    'gu': 'gujarati',
    'ht': 'haitian creole',
    'ha': 'hausa',
    'haw': 'hawaiian',
    'iw': 'hebrew',
    'he': 'hebrew',
    'hi': 'hindi',
    'hmn': 'hmong',
    'hu': 'hungarian',
    'is': 'icelandic',
    'ig': 'igbo',
    'id': 'indonesian',
    'ga': 'irish',
    'it': 'italian',
    'ja': 'japanese',
    'jw': 'javanese',
    'kn': 'kannada',
    'kk': 'kazakh',
    'km': 'khmer',
    'ko': 'korean',
    'ku': 'kurdish (kurmanji)',
    'ky': 'kyrgyz',
    'lo': 'lao',
    'la': 'latin',
    'lv': 'latvian',
    'lt': 'lithuanian',
    'lb': 'luxembourgish',
    'mk': 'macedonian',
    'mg': 'malagasy',
    'ms': 'malay',
    'ml': 'malayalam',
    'mt': 'maltese',
    'mi': 'maori',
    'mr': 'marathi',
    'mn': 'mongolian',
    'my': 'myanmar (burmese)',
    'ne': 'nepali',
    'no': 'norwegian',
    'or': 'odia',
    'ps': 'pashto',
    'fa': 'persian',
    'pl': 'polish',
    'pt': 'portuguese',
    'pa': 'punjabi',
    'ro': 'romanian',
    'ru': 'russian',
    'sm': 'samoan',
    'gd': 'scots gaelic',
    'sr': 'serbian',
    'st': 'sesotho',
    'sn': 'shona',
    'sd': 'sindhi',
    'si': 'sinhala',
    'sk': 'slovak',
    'sl': 'slovenian',
    'so': 'somali',
    'es': 'spanish',
    'su': 'sundanese',
    'sw': 'swahili',
    'sv': 'swedish',
    'tg': 'tajik',
    'ta': 'tamil',
    'te': 'telugu',
    'th': 'thai',
    'tr': 'turkish',
    'uk': 'ukrainian',
    'ur': 'urdu',
    'ug': 'uyghur',
    'uz': 'uzbek',
    'vi': 'vietnamese',
    'cy': 'welsh',
    'xh': 'xhosa',
    'yi': 'yiddish',
    'yo': 'yoruba',
    'zu': 'zulu',
}

PRIMARY_COUNTRY = {
    'af': 'ZA',  # Afrikaans - South Africa
    'sq': 'AL',  # Albanian - Albania
    'am': 'ET',  # Amharic - Ethiopia
    'ar': 'SA',  # Arabic - Saudi Arabia
    'hy': 'AM',  # Armenian - Armenia
    'az': 'AZ',  # Azerbaijani - Azerbaijan
    'eu': 'ES',  # Basque - Spain
    'be': 'BY',  # Belarusian - Belarus
    'bn': 'BD',  # Bengali - Bangladesh
    'bs': 'BA',  # Bosnian - Bosnia and Herzegovina
    'bg': 'BG',  # Bulgarian - Bulgaria
    'ca': 'ES',  # Catalan - Spain
    'ceb': 'PH',  # Cebuano - Philippines
    'ny': 'MW',  # Chichewa - Malawi
    'zh-cn': 'CN',  # Chinese (Simplified) - China
    'zh-tw': 'TW',  # Chinese (Traditional) - Taiwan
    'co': 'FR',  # Corsican - France
    'hr': 'HR',  # Croatian - Croatia
    'cs': 'CZ',  # Czech - Czech Republic
    'da': 'DK',  # Danish - Denmark
    'nl': 'NL',  # Dutch - Netherlands
    'en': 'US',  # English - United States
    'eo': None,  # Esperanto - No primary country
    'et': 'EE',  # Estonian - Estonia
    'tl': 'PH',  # Filipino - Philippines
    'fi': 'FI',  # Finnish - Finland
    'fr': 'FR',  # French - France
    'fy': 'NL',  # Frisian - Netherlands
    'gl': 'ES',  # Galician - Spain
    'ka': 'GE',  # Georgian - Georgia
    'de': 'DE',  # German - Germany
    'el': 'GR',  # Greek - Greece
    'gu': 'IN',  # Gujarati - India
    'ht': 'HT',  # Haitian Creole - Haiti
    'ha': 'NG',  # Hausa - Nigeria
    'haw': 'US',  # Hawaiian - United States (Hawaii)
    'iw': 'IL',  # Hebrew - Israel
    'he': 'IL',  # Hebrew - Israel
    'hi': 'IN',  # Hindi - India
    'hmn': 'LA',  # Hmong - Laos
    'hu': 'HU',  # Hungarian - Hungary
    'is': 'IS',  # Icelandic - Iceland
    'ig': 'NG',  # Igbo - Nigeria
    'id': 'ID',  # Indonesian - Indonesia
    'ga': 'IE',  # Irish - Ireland
    'it': 'IT',  # Italian - Italy
    'ja': 'JP',  # Japanese - Japan
    'jw': 'ID',  # Javanese - Indonesia
    'kn': 'IN',  # Kannada - India
    'kk': 'KZ',  # Kazakh - Kazakhstan
    'km': 'KH',  # Khmer - Cambodia
    'ko': 'KR',  # Korean - South Korea
    'ku': 'IQ',  # Kurdish (Kurmanji) - Iraq
    'ky': 'KG',  # Kyrgyz - Kyrgyzstan
    'lo': 'LA',  # Lao - Laos
    'la': 'VA',  # Latin - Vatican City
    'lv': 'LV',  # Latvian - Latvia
    'lt': 'LT',  # Lithuanian - Lithuania
    'lb': 'LU',  # Luxembourgish - Luxembourg
    'mk': 'MK',  # Macedonian - North Macedonia
    'mg': 'MG',  # Malagasy - Madagascar
    'ms': 'MY',  # Malay - Malaysia
    'ml': 'IN',  # Malayalam - India
    'mt': 'MT',  # Maltese - Malta
    'mi': 'NZ',  # Maori - New Zealand
    'mr': 'IN',  # Marathi - India
    'mn': 'MN',  # Mongolian - Mongolia
    'my': 'MM',  # Myanmar (Burmese) - Myanmar
    'ne': 'NP',  # Nepali - Nepal
    'no': 'NO',  # Norwegian - Norway
    'or': 'IN',  # Odia - India
    'ps': 'AF',  # Pashto - Afghanistan
    'fa': 'IR',  # Persian - Iran
    'pl': 'PL',  # Polish - Poland
    'pt': 'BR',  # Portuguese - Brazil
    'pa': 'IN',  # Punjabi - India
    'ro': 'RO',  # Romanian - Romania
    'ru': 'RU',  # Russian - Russia
    'sm': 'WS',  # Samoan - Samoa
    'gd': 'GB',  # Scots Gaelic - United Kingdom (Scotland)
    'sr': 'RS',  # Serbian - Serbia
    'st': 'LS',  # Sesotho - Lesotho
    'sn': 'ZW',  # Shona - Zimbabwe
    'sd': 'PK',  # Sindhi - Pakistan
    'si': 'LK',  # Sinhala - Sri Lanka
    'sk': 'SK',  # Slovak - Slovakia
    'sl': 'SI',  # Slovenian - Slovenia
    'so': 'SO',  # Somali - Somalia
    'es': 'ES',  # Spanish - Spain
    'su': 'ID',  # Sundanese - Indonesia
    'sw': 'KE',  # Swahili - Kenya
    'sv': 'SE',  # Swedish - Sweden
    'tg': 'TJ',  # Tajik - Tajikistan
    'ta': 'IN',  # Tamil - India
    'te': 'IN',  # Telugu - India
    'th': 'TH',  # Thai - Thailand
    'tr': 'TR',  # Turkish - Turkey
    'uk': 'UA',  # Ukrainian - Ukraine
    'ur': 'PK',  # Urdu - Pakistan
    'ug': 'CN',  # Uyghur - China
    'uz': 'UZ',  # Uzbek - Uzbekistan
    'vi': 'VN',  # Vietnamese - Vietnam
    'cy': 'GB',  # Welsh - United Kingdom (Wales)
    'xh': 'ZA',  # Xhosa - South Africa
    'yi': None,  # Yiddish - No primary country
    'yo': 'NG',  # Yoruba - Nigeria
    'zu': 'ZA'  # Zulu - South Africa
}

LANGCODES = dict(map(reversed, LANGUAGES.items()))
DEFAULT_RAISE_EXCEPTION = False
DUMMY_DATA = [[["", None, None, 0]], None, "en", None,
              None, None, 1, None, [["en"], None, [1], ["en"]]]



# -*- coding: utf-8 -*-
"""
An Async Translation module.

You can translate text using this module, asynchronously.
"""
import httpx

from idioma import urls, utils
from idioma.async_gtoken import AsyncTokenAcquirer
from idioma.base_translator import BaseTranslator
from idioma.exceptions import EmptyTranslationData
from idioma.models import Detected
from idioma.retry_transport import RetryTransport


class AsyncTranslator(BaseTranslator):

    def _create_client(self) -> httpx.AsyncClient:
        # wrapped transport ensures that the client will retry on failure
        # of connections, not based on status codes
        # retry transport is used to retry on status codes
        retry_transport = RetryTransport(
            wrapped_transport=httpx.AsyncHTTPTransport(retries=10),
            retryable_methods={'GET', 'POST'},
            # Google would take too "sorry" page sometimes with 302
            retry_status_codes={302}
        )
        return httpx.AsyncClient(
            http2=self.http2,
            proxies=self.proxies,
            timeout=self.timeout,
            transport=retry_transport
        )

    def _get_token_acquirer(self):
        return AsyncTokenAcquirer(client=self.client,
                                  host=self.service_urls[0])

    async def prepare_translate_legacy_params(self, text, src, dest,
                                              override=None):
        token = ''  # dummy default value here as it is not used by api client
        if self.client_type == 'webapp':
            token = await self.token_acquirer.do(text)

        params = utils.build_params(client=self.client_type, query=text,
                                    src=src, dest=dest,
                                    token=token, override=override)

        url = urls.TRANSLATE.format(host=self._pick_service_url())
        return url, params

    async def _translate(self, text: str, dest: str, src: str):
        url = urls.TRANSLATE_RPC.format(host=self._pick_service_url())
        data = {
            'f.req': self._build_rpc_request(text, dest, src),
        }
        r = await self.client.post(url, params=self.POST_PARAMS, data=data)

        if r.status_code != 200 and self.raise_exception:
            raise Exception('Unexpected status code "{}" from {}'.format(
                r.status_code, self.service_urls))

        return r.text, r

    async def _translate_legacy(self, text, dest, src, override):

        url, params = await self.prepare_translate_legacy_params(text, src,
                                                                 dest,
                                                                 override)
        response = await self.client.get(url, params=params)
        return self.handle_legacy_translate_response(response, text)

    async def translate(self, text: str, dest='en', src=None):
        src, dest = self.validate_normalize_src_dest(src, dest)
        data, response = await self._translate(text, dest, src)
        try:
            return self.parse_response(src, dest, text, data, response)
        except EmptyTranslationData:
            return await self.translate_legacy(text, dest=dest, src=src)

    async def detect(self, text: str):
        translated = await self.translate(text, src='auto', dest='en')
        result = Detected(lang=translated.src,
                          confidence=translated.extra_data.get('confidence',
                                                               None),
                          response=translated._response)
        return result

    def detect_legacy(self, text, **kwargs):
        """Detect language of the input text

        :param text: The source text(s) whose language you want to identify.
                     Batch detection is supported via sequence input.
        :type text: UTF-8 :class:`str`; :class:`unicode`; string sequence (list, tuple, iterator, generator)

        :rtype: Detected
        :rtype: :class:`list` (when a list is passed)

        Basic usage:
            >>> from idioma import AsyncTranslator
            >>> translator = AsyncTranslator()
            >>>  translator.detect('이 문장은 한글로 쓰여졌습니다.')
            <Detected lang=ko confidence=0.27041003>
            >>> translator.detect('この文章は日本語で書かれました。')
            <Detected lang=ja confidence=0.64889508>
            >>> translator.detect('This sentence is written in English.')
            <Detected lang=en confidence=0.22348526>
            >>> translator.detect('Tiu frazo estas skribita en Esperanto.')
            <Detected lang=eo confidence=0.10538048>

        Advanced usage:
            >>> langs = translator.detect(['한국어', '日本語', 'English', 'le français'])
            >>> for lang in langs:
            ...    print(lang.lang, lang.confidence)
            ko 1
            ja 0.92929292
            en 0.96954316
            fr 0.043500196
        """
        if isinstance(text, list):
            result = []
            for item in text:
                lang = self.detect(item)
                result.append(lang)
            return result

        data, response = self._translate_legacy(text, 'en', 'auto', kwargs)

        return self.extract_source_language_and_confidence(data, response)

    async def __aenter__(self):
        # Initialize the httpx.AsyncClient here
        self.client = self._create_client()

        if self.proxies is not None:
            self.client.proxies = self.proxies

        self.client.headers.update({
            'User-Agent': self.user_agent,
            'Referer': 'https://translate.google.com',
        })

        if self.timeout is not None:
            self.client.timeout = self.timeout

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        # Close the httpx.AsyncClient when exiting the context
        await self.client.aclose()

    async def translate_legacy(self, text, dest='en', src='auto', **kwargs):
        """Translate text from source language to destination language

        :param text: The source text(s) to be translated. Batch translation is supported via sequence input.
        :type text: UTF-8 :class:`str`; :class:`unicode`; string sequence (list, tuple, iterator, generator)

        :param dest: The language to translate the source text into.
                     The value should be one of the language codes listed in :const:`idioma.LANGUAGES`
                     or one of the language names listed in :const:`idioma.LANGCODES`.
        :param dest: :class:`str`; :class:`unicode`

        :param src: The language of the source text.
                    The value should be one of the language codes listed in :const:`idioma.LANGUAGES`
                    or one of the language names listed in :const:`idioma.LANGCODES`.
                    If a language is not specified,
                    the system will attempt to identify the source language automatically.
        :param src: :class:`str`; :class:`unicode`

        :rtype: Translated
        :rtype: :class:`list` (when a list is passed)

        Basic usage:
            >>> from idioma import Translator
            >>> translator = Translator()
            >>> translator.translate('안녕하세요.')
            <Translated src=ko dest=en text=Good evening. pronunciation=Good evening.>
            >>> translator.translate('안녕하세요.', dest='ja')
            <Translated src=ko dest=ja text=こんにちは。 pronunciation=Kon'nichiwa.>
            >>> translator.translate('veritas lux mea', src='la')
            <Translated src=la dest=en text=The truth is my light pronunciation=The truth is my light>

        Advanced usage:
            >>> translations = translator.translate(['The quick brown fox', 'jumps over', 'the lazy dog'], dest='ko')
            >>> for translation in translations:
            ...    print(translation.origin, ' -> ', translation.text)
            The quick brown fox  ->  빠른 갈색 여우
            jumps over  ->  이상 점프
            the lazy dog  ->  게으른 개
        """
        src, dest = self.validate_normalize_src_dest(src, dest)

        if isinstance(text, list):
            result = []
            for item in text:
                translated = await self.translate_legacy(item, dest=dest,
                                                         src=src,
                                                         **kwargs)
                result.append(translated)
            return result

        data, response = self._translate_legacy(text, dest, src, kwargs)

        return self.parse_legacy_response(data, src, dest, text, response)



import asyncio
import time

import pytest

from idioma import AsyncTranslator, Translator

# Number of translations to perform for the test
NUM_TRANSLATIONS = 10  # Adjust this to a small number for the test


@pytest.mark.asyncio
async def test_async_translation():
    async with AsyncTranslator() as translator:
        text_to_translate = "Hello, world!"
        destination_language = "fr"
        translation = await translator.translate(text_to_translate,
                                                 dest=destination_language)
        assert translation is not None
        assert translation.src == "en"  # Ensure the source language is correctly detected
        assert translation.dest == destination_language
        assert translation.text is not None
        assert translation.text == "Bonjour le monde!"  # Check that translation is different from the input text


@pytest.mark.asyncio
async def test_async_translate_thai_lang_detect():
    async with AsyncTranslator() as translator:
        detection = await translator.detect('안녕하세요')
        assert detection.lang == 'ko'


@pytest.mark.asyncio
async def test_async_translate_thai_explicit():
    async with AsyncTranslator() as translator:
        translation = await translator.translate(
            'แต่ส่งได้แค่คำว่าทำไม', src='th', dest='en'
        )
        assert translation.text == 'But can only send the word'


@pytest.mark.asyncio
async def test_async_is_faster_than_sync():
    translator = Translator()
    async_translator = AsyncTranslator()

    async def translate_sync():
        for _ in range(NUM_TRANSLATIONS):
            translation = translator.translate("Hello, world!", dest='fr')
            assert translation.text == "Bonjour le monde!"

    async def translate_async():
        tasks = []
        for _ in range(NUM_TRANSLATIONS):
            task = async_translator.translate("Hello, world!", dest='fr')
            tasks.append(task)
        await asyncio.gather(*tasks)

    sync_start_time = time.time()
    await translate_sync()
    sync_end_time = time.time()
    sync_elapsed_time = sync_end_time - sync_start_time

    async_start_time = time.time()
    await translate_async()
    async_end_time = time.time()
    async_elapsed_time = async_end_time - async_start_time

    # Assert that the asynchronous version is faster than the synchronous
    # version
    assert async_elapsed_time < sync_elapsed_time


@pytest.mark.asyncio
async def test_translate_russian_lang_hello_evade_rate_limit(capsys):
    translator = AsyncTranslator()
    for i in range(300):
        translation = await translator.translate(
            f'Привет #{i}', src='ru', dest='en')
        translated = translation.text
        print(f"Translated: {translated}")
        assert translated in [f'Hello #{i}', f'Hi #{i}', f'Hi: #{i}', f"Hi: {i}"]



# -*- coding: utf-8 -*-

from idioma.base_gtoken import BaseTokenAcquirer


class AsyncTokenAcquirer(BaseTokenAcquirer):
    async def _update(self):
        """update tkk
        """
        if not self._get_current_time():
            return

        r = await self.client.get(self.host)

        self._set_tkk_from_response(r)

    async def do(self, text):
        await self._update()
        tk = self.acquire(text)
        return tk
